/*
 * Copyright (C) Cross The Road Electronics.  All rights reserved.
 * License information can be found in CTRE_LICENSE.txt
 * For support and suggestions contact support@ctr-electronics.com or file
 * an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
 */
#pragma once

#include "ctre/phoenix6/core/CoreCANcoder.hpp"
#include "ctre/phoenix6/core/CoreTalonFX.hpp"

#include <units/length.h>
#include <units/moment_of_inertia.h>
#include <units/velocity.h>

namespace ctre {
namespace phoenix6 {
namespace swerve {

/**
 * \brief Supported closed-loop output types.
 */
enum class ClosedLoopOutputType {
    Voltage = 0,
    /** \brief Requires Pro */
    TorqueCurrentFOC = 1,
};

/**
 * \brief Supported feedback sensors for the steer motors.
 */
enum class SteerFeedbackType {
    /**
     * \brief Requires Pro; Use signals#FeedbackSensorSourceValue#FusedCANcoder
     * for the steer motor.
     */
    FusedCANcoder = 0,
    /**
     * \brief Requires Pro; Use signals#FeedbackSensorSourceValue#SyncCANcoder
     * for the steer motor.
     */
    SyncCANcoder = 1,
    /**
     * \brief Use signals#FeedbackSensorSourceValue#RemoteCANcoder
     * for the steer motor.
     */
    RemoteCANcoder = 2,
};

/**
 * \brief All constants for a swerve module.
 */
struct SwerveModuleConstants {
    constexpr SwerveModuleConstants() = default;

    /**
     * \brief CAN ID of the steer motor.
     */
    int SteerMotorId = 0;
    /**
     * \brief CAN ID of the drive motor.
     */
    int DriveMotorId = 0;
    /**
     * \brief CAN ID of the CANcoder used for azimuth.
     */
    int CANcoderId = 0;
    /**
     * \brief Offset of the CANcoder.
     */
    units::angle::turn_t CANcoderOffset = 0_tr;
    /**
     * \brief The location of this module's wheels relative to the physical center
     * of the robot in meters along the X axis of the robot.
     */
    units::length::meter_t LocationX = 0_m;
    /**
     * \brief The location of this module's wheels relative to the physical center
     * of the robot in meters along the Y axis of the robot.
     */
    units::length::meter_t LocationY = 0_m;
    /**
     * \brief True if the drive motor is inverted.
     */
    bool DriveMotorInverted = false;
    /**
     * \brief True if the steer motor is inverted from the CANcoder.
     */
    bool SteerMotorInverted = false;
    /**
     * \brief Gear ratio between the drive motor and the wheel.
     */
    units::dimensionless::scalar_t DriveMotorGearRatio = 0;
    /**
     * \brief Gear ratio between the steer motor and the CANcoder. For example, the
     * SDS Mk4 has a steering ratio of 12.8.
     */
    units::dimensionless::scalar_t SteerMotorGearRatio = 0;
    /**
     * \brief Coupled gear ratio between the CANcoder and the drive motor.
     * 
     * For a typical swerve module, the azimuth turn motor also drives the wheel a
     * nontrivial amount, which affects the accuracy of odometry and control. This
     * ratio represents the number of rotations of the drive motor caused by a
     * rotation of the azimuth.
     */
    units::dimensionless::scalar_t CouplingGearRatio = 0;
    /**
     * \brief Radius of the driving wheel in meters.
     */
    units::length::meter_t WheelRadius = 0_m;
    /**
     * \brief The steer motor closed-loop gains.
     * 
     * The steer motor uses the control ouput type specified by
     * SteerMotorClosedLoopOutput and any SwerveModule#SteerRequestType.
     */
    configs::Slot0Configs SteerMotorGains = configs::Slot0Configs{};
    /**
     * \brief The drive motor closed-loop gains.
     * 
     * When using closed-loop control, the drive motor uses the control output type
     * specified by DriveMotorClosedLoopOutput and any closed-loop
     * SwerveModule#DriveRequestType.
     */
    configs::Slot0Configs DriveMotorGains = configs::Slot0Configs{};
    /**
     * \brief The closed-loop output type to use for the steer motors.
     */
    ClosedLoopOutputType SteerMotorClosedLoopOutput = ClosedLoopOutputType::Voltage;
    /**
     * \brief The closed-loop output type to use for the drive motors.
     */
    ClosedLoopOutputType DriveMotorClosedLoopOutput = ClosedLoopOutputType::Voltage;
    /**
     * \brief The maximum amount of stator current the drive motors can apply
     * without slippage.
     */
    units::current::ampere_t SlipCurrent = 120_A;
    /**
     * \brief When using open-loop drive control, this specifies the speed at which
     * the robot travels when driven with 12 volts. This is used to approximate the
     * output for a desired velocity. If using closed loop control, this value is
     * ignored.
     */
    units::velocity::meters_per_second_t SpeedAt12Volts = 0_mps;
    /**
     * \brief Choose how the feedback sensors should be configured.
     * 
     * If the robot does not support Pro, then this should be set to RemoteCANcoder.
     * Otherwise, users have the option to use either FusedCANcoder or SyncCANcoder
     * depending on if there is a risk that the CANcoder can fail in a way to
     * provide "good" data.
     * 
     * If this is set to FusedCANcoder or SyncCANcoder when the steer motor is not
     * Pro-licensed, the device will automatically fall back to RemoteCANcoder and
     * report a UsingProFeatureOnUnlicensedDevice status code.
     */
    SteerFeedbackType FeedbackSource = SteerFeedbackType::FusedCANcoder;
    /**
     * \brief The initial configs used to configure the drive motor of the swerve
     * module. The default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MotorOutputConfigs#NeutralMode (Brake mode, overwritten with
     *   SwerveDrivetrain#ConfigNeutralMode)
     * - configs#MotorOutputConfigs#Inverted
     *   (SwerveModuleConstants#DriveMotorInverted)
     * - configs#Slot0Configs (SwerveModuleConstants#DriveMotorGains)
     * - configs#CurrentLimitsConfigs#StatorCurrentLimit /
     *   configs#TorqueCurrentConfigs#PeakForwardTorqueCurrent /
     *   configs#TorqueCurrentConfigs#PeakReverseTorqueCurrent
     *   (SwerveModuleConstants#SlipCurrent)
     * - configs#CurrentLimitsConfigs#StatorCurrentLimitEnable (Enabled)
     * 
     */
    configs::TalonFXConfiguration DriveMotorInitialConfigs = configs::TalonFXConfiguration{};
    /**
     * \brief The initial configs used to configure the steer motor of the swerve
     * module. The default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MotorOutputConfigs#NeutralMode (Brake mode)
     * - configs#MotorOutputConfigs#Inverted
     *   (SwerveModuleConstants#SteerMotorInverted)
     * - configs#Slot0Configs (SwerveModuleConstants#SteerMotorGains)
     * - configs#FeedbackConfigs#FeedbackRemoteSensorID
     *   (SwerveModuleConstants#CANcoderId)
     * - configs#FeedbackConfigs#FeedbackSensorSource
     *   (SwerveModuleConstants#FeedbackSource)
     * - configs#FeedbackConfigs#RotorToSensorRatio
     *   (SwerveModuleConstants#SteerMotorGearRatio)
     * - configs#MotionMagicConfigs (Calculated from gear ratios)
     * - configs#ClosedLoopGeneralConfigs#ContinuousWrap (true)
     * 
     */
    configs::TalonFXConfiguration SteerMotorInitialConfigs = configs::TalonFXConfiguration{};
    /**
     * \brief The initial configs used to configure the CANcoder of the swerve
     * module. The default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MagnetSensorConfigs#MagnetOffset
     *   (SwerveModuleConstants#CANcoderOffset)
     * 
     */
    configs::CANcoderConfiguration CANcoderInitialConfigs = configs::CANcoderConfiguration{};
    /**
     * \brief Simulated azimuthal inertia.
     */
    units::moment_of_inertia::kilogram_square_meter_t SteerInertia = 0.00001_kg_sq_m;
    /**
     * \brief Simulated drive inertia.
     */
    units::moment_of_inertia::kilogram_square_meter_t DriveInertia = 0.001_kg_sq_m;
    /**
     * \brief Simulated steer voltage required to overcome friction.
     */
    units::voltage::volt_t SteerFrictionVoltage = 0.25_V;
    /**
     * \brief Simulated drive voltage required to overcome friction.
     */
    units::voltage::volt_t DriveFrictionVoltage = 0.25_V;
    
    /**
     * \brief Modifies the SteerMotorId parameter and returns itself.
     *
     * CAN ID of the steer motor.
     *
     * \param newSteerMotorId Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithSteerMotorId(int newSteerMotorId)
    {
        this->SteerMotorId = newSteerMotorId;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveMotorId parameter and returns itself.
     *
     * CAN ID of the drive motor.
     *
     * \param newDriveMotorId Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithDriveMotorId(int newDriveMotorId)
    {
        this->DriveMotorId = newDriveMotorId;
        return *this;
    }
    
    /**
     * \brief Modifies the CANcoderId parameter and returns itself.
     *
     * CAN ID of the CANcoder used for azimuth.
     *
     * \param newCANcoderId Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithCANcoderId(int newCANcoderId)
    {
        this->CANcoderId = newCANcoderId;
        return *this;
    }
    
    /**
     * \brief Modifies the CANcoderOffset parameter and returns itself.
     *
     * Offset of the CANcoder.
     *
     * \param newCANcoderOffset Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithCANcoderOffset(units::angle::turn_t newCANcoderOffset)
    {
        this->CANcoderOffset = newCANcoderOffset;
        return *this;
    }
    
    /**
     * \brief Modifies the LocationX parameter and returns itself.
     *
     * The location of this module's wheels relative to the physical center of the
     * robot in meters along the X axis of the robot.
     *
     * \param newLocationX Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithLocationX(units::length::meter_t newLocationX)
    {
        this->LocationX = newLocationX;
        return *this;
    }
    
    /**
     * \brief Modifies the LocationY parameter and returns itself.
     *
     * The location of this module's wheels relative to the physical center of the
     * robot in meters along the Y axis of the robot.
     *
     * \param newLocationY Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithLocationY(units::length::meter_t newLocationY)
    {
        this->LocationY = newLocationY;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveMotorInverted parameter and returns itself.
     *
     * True if the drive motor is inverted.
     *
     * \param newDriveMotorInverted Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithDriveMotorInverted(bool newDriveMotorInverted)
    {
        this->DriveMotorInverted = newDriveMotorInverted;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerMotorInverted parameter and returns itself.
     *
     * True if the steer motor is inverted from the CANcoder.
     *
     * \param newSteerMotorInverted Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithSteerMotorInverted(bool newSteerMotorInverted)
    {
        this->SteerMotorInverted = newSteerMotorInverted;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveMotorGearRatio parameter and returns itself.
     *
     * Gear ratio between the drive motor and the wheel.
     *
     * \param newDriveMotorGearRatio Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithDriveMotorGearRatio(units::dimensionless::scalar_t newDriveMotorGearRatio)
    {
        this->DriveMotorGearRatio = newDriveMotorGearRatio;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerMotorGearRatio parameter and returns itself.
     *
     * Gear ratio between the steer motor and the CANcoder. For example, the SDS Mk4
     * has a steering ratio of 12.8.
     *
     * \param newSteerMotorGearRatio Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithSteerMotorGearRatio(units::dimensionless::scalar_t newSteerMotorGearRatio)
    {
        this->SteerMotorGearRatio = newSteerMotorGearRatio;
        return *this;
    }
    
    /**
     * \brief Modifies the CouplingGearRatio parameter and returns itself.
     *
     * Coupled gear ratio between the CANcoder and the drive motor.
     * 
     * For a typical swerve module, the azimuth turn motor also drives the wheel a
     * nontrivial amount, which affects the accuracy of odometry and control. This
     * ratio represents the number of rotations of the drive motor caused by a
     * rotation of the azimuth.
     *
     * \param newCouplingGearRatio Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithCouplingGearRatio(units::dimensionless::scalar_t newCouplingGearRatio)
    {
        this->CouplingGearRatio = newCouplingGearRatio;
        return *this;
    }
    
    /**
     * \brief Modifies the WheelRadius parameter and returns itself.
     *
     * Radius of the driving wheel in meters.
     *
     * \param newWheelRadius Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithWheelRadius(units::length::meter_t newWheelRadius)
    {
        this->WheelRadius = newWheelRadius;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerMotorGains parameter and returns itself.
     *
     * The steer motor closed-loop gains.
     * 
     * The steer motor uses the control ouput type specified by
     * SteerMotorClosedLoopOutput and any SwerveModule#SteerRequestType.
     *
     * \param newSteerMotorGains Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithSteerMotorGains(const configs::Slot0Configs& newSteerMotorGains)
    {
        this->SteerMotorGains = newSteerMotorGains;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveMotorGains parameter and returns itself.
     *
     * The drive motor closed-loop gains.
     * 
     * When using closed-loop control, the drive motor uses the control output type
     * specified by DriveMotorClosedLoopOutput and any closed-loop
     * SwerveModule#DriveRequestType.
     *
     * \param newDriveMotorGains Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithDriveMotorGains(const configs::Slot0Configs& newDriveMotorGains)
    {
        this->DriveMotorGains = newDriveMotorGains;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerMotorClosedLoopOutput parameter and returns itself.
     *
     * The closed-loop output type to use for the steer motors.
     *
     * \param newSteerMotorClosedLoopOutput Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithSteerMotorClosedLoopOutput(ClosedLoopOutputType newSteerMotorClosedLoopOutput)
    {
        this->SteerMotorClosedLoopOutput = newSteerMotorClosedLoopOutput;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveMotorClosedLoopOutput parameter and returns itself.
     *
     * The closed-loop output type to use for the drive motors.
     *
     * \param newDriveMotorClosedLoopOutput Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithDriveMotorClosedLoopOutput(ClosedLoopOutputType newDriveMotorClosedLoopOutput)
    {
        this->DriveMotorClosedLoopOutput = newDriveMotorClosedLoopOutput;
        return *this;
    }
    
    /**
     * \brief Modifies the SlipCurrent parameter and returns itself.
     *
     * The maximum amount of stator current the drive motors can apply without
     * slippage.
     *
     * \param newSlipCurrent Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithSlipCurrent(units::current::ampere_t newSlipCurrent)
    {
        this->SlipCurrent = newSlipCurrent;
        return *this;
    }
    
    /**
     * \brief Modifies the SpeedAt12Volts parameter and returns itself.
     *
     * When using open-loop drive control, this specifies the speed at which the
     * robot travels when driven with 12 volts. This is used to approximate the
     * output for a desired velocity. If using closed loop control, this value is
     * ignored.
     *
     * \param newSpeedAt12Volts Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithSpeedAt12Volts(units::velocity::meters_per_second_t newSpeedAt12Volts)
    {
        this->SpeedAt12Volts = newSpeedAt12Volts;
        return *this;
    }
    
    /**
     * \brief Modifies the FeedbackSource parameter and returns itself.
     *
     * Choose how the feedback sensors should be configured.
     * 
     * If the robot does not support Pro, then this should be set to RemoteCANcoder.
     * Otherwise, users have the option to use either FusedCANcoder or SyncCANcoder
     * depending on if there is a risk that the CANcoder can fail in a way to
     * provide "good" data.
     * 
     * If this is set to FusedCANcoder or SyncCANcoder when the steer motor is not
     * Pro-licensed, the device will automatically fall back to RemoteCANcoder and
     * report a UsingProFeatureOnUnlicensedDevice status code.
     *
     * \param newFeedbackSource Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithFeedbackSource(SteerFeedbackType newFeedbackSource)
    {
        this->FeedbackSource = newFeedbackSource;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveMotorInitialConfigs parameter and returns itself.
     *
     * The initial configs used to configure the drive motor of the swerve module.
     * The default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MotorOutputConfigs#NeutralMode (Brake mode, overwritten with
     *   SwerveDrivetrain#ConfigNeutralMode)
     * - configs#MotorOutputConfigs#Inverted
     *   (SwerveModuleConstants#DriveMotorInverted)
     * - configs#Slot0Configs (SwerveModuleConstants#DriveMotorGains)
     * - configs#CurrentLimitsConfigs#StatorCurrentLimit /
     *   configs#TorqueCurrentConfigs#PeakForwardTorqueCurrent /
     *   configs#TorqueCurrentConfigs#PeakReverseTorqueCurrent
     *   (SwerveModuleConstants#SlipCurrent)
     * - configs#CurrentLimitsConfigs#StatorCurrentLimitEnable (Enabled)
     * 
     *
     * \param newDriveMotorInitialConfigs Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithDriveMotorInitialConfigs(const configs::TalonFXConfiguration& newDriveMotorInitialConfigs)
    {
        this->DriveMotorInitialConfigs = newDriveMotorInitialConfigs;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerMotorInitialConfigs parameter and returns itself.
     *
     * The initial configs used to configure the steer motor of the swerve module.
     * The default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MotorOutputConfigs#NeutralMode (Brake mode)
     * - configs#MotorOutputConfigs#Inverted
     *   (SwerveModuleConstants#SteerMotorInverted)
     * - configs#Slot0Configs (SwerveModuleConstants#SteerMotorGains)
     * - configs#FeedbackConfigs#FeedbackRemoteSensorID
     *   (SwerveModuleConstants#CANcoderId)
     * - configs#FeedbackConfigs#FeedbackSensorSource
     *   (SwerveModuleConstants#FeedbackSource)
     * - configs#FeedbackConfigs#RotorToSensorRatio
     *   (SwerveModuleConstants#SteerMotorGearRatio)
     * - configs#MotionMagicConfigs (Calculated from gear ratios)
     * - configs#ClosedLoopGeneralConfigs#ContinuousWrap (true)
     * 
     *
     * \param newSteerMotorInitialConfigs Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithSteerMotorInitialConfigs(const configs::TalonFXConfiguration& newSteerMotorInitialConfigs)
    {
        this->SteerMotorInitialConfigs = newSteerMotorInitialConfigs;
        return *this;
    }
    
    /**
     * \brief Modifies the CANcoderInitialConfigs parameter and returns itself.
     *
     * The initial configs used to configure the CANcoder of the swerve module. The
     * default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MagnetSensorConfigs#MagnetOffset
     *   (SwerveModuleConstants#CANcoderOffset)
     * 
     *
     * \param newCANcoderInitialConfigs Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithCANcoderInitialConfigs(const configs::CANcoderConfiguration& newCANcoderInitialConfigs)
    {
        this->CANcoderInitialConfigs = newCANcoderInitialConfigs;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerInertia parameter and returns itself.
     *
     * Simulated azimuthal inertia.
     *
     * \param newSteerInertia Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithSteerInertia(units::moment_of_inertia::kilogram_square_meter_t newSteerInertia)
    {
        this->SteerInertia = newSteerInertia;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveInertia parameter and returns itself.
     *
     * Simulated drive inertia.
     *
     * \param newDriveInertia Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithDriveInertia(units::moment_of_inertia::kilogram_square_meter_t newDriveInertia)
    {
        this->DriveInertia = newDriveInertia;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerFrictionVoltage parameter and returns itself.
     *
     * Simulated steer voltage required to overcome friction.
     *
     * \param newSteerFrictionVoltage Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithSteerFrictionVoltage(units::voltage::volt_t newSteerFrictionVoltage)
    {
        this->SteerFrictionVoltage = newSteerFrictionVoltage;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveFrictionVoltage parameter and returns itself.
     *
     * Simulated drive voltage required to overcome friction.
     *
     * \param newDriveFrictionVoltage Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstants &WithDriveFrictionVoltage(units::voltage::volt_t newDriveFrictionVoltage)
    {
        this->DriveFrictionVoltage = newDriveFrictionVoltage;
        return *this;
    }
    
};

/**
 * \brief Constants that are common across the swerve modules, used
 * for creating instances of module-specific SwerveModuleConstants.
 */
struct SwerveModuleConstantsFactory {
    constexpr SwerveModuleConstantsFactory() = default;

    /**
     * \brief Gear ratio between the drive motor and the wheel.
     */
    units::dimensionless::scalar_t DriveMotorGearRatio = 0;
    /**
     * \brief Gear ratio between the steer motor and the CANcoder. For example, the
     * SDS Mk4 has a steering ratio of 12.8.
     */
    units::dimensionless::scalar_t SteerMotorGearRatio = 0;
    /**
     * \brief Coupled gear ratio between the CANcoder and the drive motor.
     * 
     * For a typical swerve module, the azimuth turn motor also drives the wheel a
     * nontrivial amount, which affects the accuracy of odometry and control. This
     * ratio represents the number of rotations of the drive motor caused by a
     * rotation of the azimuth.
     */
    units::dimensionless::scalar_t CouplingGearRatio = 0;
    /**
     * \brief Radius of the driving wheel in meters.
     */
    units::length::meter_t WheelRadius = 0_m;
    /**
     * \brief The steer motor closed-loop gains.
     * 
     * The steer motor uses the control ouput type specified by
     * SteerMotorClosedLoopOutput and any SwerveModule#SteerRequestType.
     */
    configs::Slot0Configs SteerMotorGains = configs::Slot0Configs{};
    /**
     * \brief The drive motor closed-loop gains.
     * 
     * When using closed-loop control, the drive motor uses the control output type
     * specified by DriveMotorClosedLoopOutput and any closed-loop
     * SwerveModule#DriveRequestType.
     */
    configs::Slot0Configs DriveMotorGains = configs::Slot0Configs{};
    /**
     * \brief The closed-loop output type to use for the steer motors.
     */
    ClosedLoopOutputType SteerMotorClosedLoopOutput = ClosedLoopOutputType::Voltage;
    /**
     * \brief The closed-loop output type to use for the drive motors.
     */
    ClosedLoopOutputType DriveMotorClosedLoopOutput = ClosedLoopOutputType::Voltage;
    /**
     * \brief The maximum amount of stator current the drive motors can apply
     * without slippage.
     */
    units::current::ampere_t SlipCurrent = 120_A;
    /**
     * \brief When using open-loop drive control, this specifies the speed at which
     * the robot travels when driven with 12 volts. This is used to approximate the
     * output for a desired velocity. If using closed loop control, this value is
     * ignored.
     */
    units::velocity::meters_per_second_t SpeedAt12Volts = 0_mps;
    /**
     * \brief Choose how the feedback sensors should be configured.
     * 
     * If the robot does not support Pro, then this should be set to RemoteCANcoder.
     * Otherwise, users have the option to use either FusedCANcoder or SyncCANcoder
     * depending on if there is a risk that the CANcoder can fail in a way to
     * provide "good" data.
     * 
     * If this is set to FusedCANcoder or SyncCANcoder when the steer motor is not
     * Pro-licensed, the device will automatically fall back to RemoteCANcoder and
     * report a UsingProFeatureOnUnlicensedDevice status code.
     */
    SteerFeedbackType FeedbackSource = SteerFeedbackType::FusedCANcoder;
    /**
     * \brief The initial configs used to configure the drive motor of the swerve
     * module. The default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MotorOutputConfigs#NeutralMode (Brake mode, overwritten with
     *   SwerveDrivetrain#ConfigNeutralMode)
     * - configs#MotorOutputConfigs#Inverted
     *   (SwerveModuleConstants#DriveMotorInverted)
     * - configs#Slot0Configs (SwerveModuleConstants#DriveMotorGains)
     * - configs#CurrentLimitsConfigs#StatorCurrentLimit /
     *   configs#TorqueCurrentConfigs#PeakForwardTorqueCurrent /
     *   configs#TorqueCurrentConfigs#PeakReverseTorqueCurrent
     *   (SwerveModuleConstants#SlipCurrent)
     * - configs#CurrentLimitsConfigs#StatorCurrentLimitEnable (Enabled)
     * 
     */
    configs::TalonFXConfiguration DriveMotorInitialConfigs = configs::TalonFXConfiguration{};
    /**
     * \brief The initial configs used to configure the steer motor of the swerve
     * module. The default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MotorOutputConfigs#NeutralMode (Brake mode)
     * - configs#MotorOutputConfigs#Inverted
     *   (SwerveModuleConstants#SteerMotorInverted)
     * - configs#Slot0Configs (SwerveModuleConstants#SteerMotorGains)
     * - configs#FeedbackConfigs#FeedbackRemoteSensorID
     *   (SwerveModuleConstants#CANcoderId)
     * - configs#FeedbackConfigs#FeedbackSensorSource
     *   (SwerveModuleConstants#FeedbackSource)
     * - configs#FeedbackConfigs#RotorToSensorRatio
     *   (SwerveModuleConstants#SteerMotorGearRatio)
     * - configs#MotionMagicConfigs (Calculated from gear ratios)
     * - configs#ClosedLoopGeneralConfigs#ContinuousWrap (true)
     * 
     */
    configs::TalonFXConfiguration SteerMotorInitialConfigs = configs::TalonFXConfiguration{};
    /**
     * \brief The initial configs used to configure the CANcoder of the swerve
     * module. The default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MagnetSensorConfigs#MagnetOffset
     *   (SwerveModuleConstants#CANcoderOffset)
     * 
     */
    configs::CANcoderConfiguration CANcoderInitialConfigs = configs::CANcoderConfiguration{};
    /**
     * \brief Simulated azimuthal inertia.
     */
    units::moment_of_inertia::kilogram_square_meter_t SteerInertia = 0.00001_kg_sq_m;
    /**
     * \brief Simulated drive inertia.
     */
    units::moment_of_inertia::kilogram_square_meter_t DriveInertia = 0.001_kg_sq_m;
    /**
     * \brief Simulated steer voltage required to overcome friction.
     */
    units::voltage::volt_t SteerFrictionVoltage = 0.25_V;
    /**
     * \brief Simulated drive voltage required to overcome friction.
     */
    units::voltage::volt_t DriveFrictionVoltage = 0.25_V;
    
    /**
     * \brief Modifies the DriveMotorGearRatio parameter and returns itself.
     *
     * Gear ratio between the drive motor and the wheel.
     *
     * \param newDriveMotorGearRatio Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithDriveMotorGearRatio(units::dimensionless::scalar_t newDriveMotorGearRatio)
    {
        this->DriveMotorGearRatio = newDriveMotorGearRatio;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerMotorGearRatio parameter and returns itself.
     *
     * Gear ratio between the steer motor and the CANcoder. For example, the SDS Mk4
     * has a steering ratio of 12.8.
     *
     * \param newSteerMotorGearRatio Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithSteerMotorGearRatio(units::dimensionless::scalar_t newSteerMotorGearRatio)
    {
        this->SteerMotorGearRatio = newSteerMotorGearRatio;
        return *this;
    }
    
    /**
     * \brief Modifies the CouplingGearRatio parameter and returns itself.
     *
     * Coupled gear ratio between the CANcoder and the drive motor.
     * 
     * For a typical swerve module, the azimuth turn motor also drives the wheel a
     * nontrivial amount, which affects the accuracy of odometry and control. This
     * ratio represents the number of rotations of the drive motor caused by a
     * rotation of the azimuth.
     *
     * \param newCouplingGearRatio Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithCouplingGearRatio(units::dimensionless::scalar_t newCouplingGearRatio)
    {
        this->CouplingGearRatio = newCouplingGearRatio;
        return *this;
    }
    
    /**
     * \brief Modifies the WheelRadius parameter and returns itself.
     *
     * Radius of the driving wheel in meters.
     *
     * \param newWheelRadius Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithWheelRadius(units::length::meter_t newWheelRadius)
    {
        this->WheelRadius = newWheelRadius;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerMotorGains parameter and returns itself.
     *
     * The steer motor closed-loop gains.
     * 
     * The steer motor uses the control ouput type specified by
     * SteerMotorClosedLoopOutput and any SwerveModule#SteerRequestType.
     *
     * \param newSteerMotorGains Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithSteerMotorGains(configs::Slot0Configs newSteerMotorGains)
    {
        this->SteerMotorGains = newSteerMotorGains;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveMotorGains parameter and returns itself.
     *
     * The drive motor closed-loop gains.
     * 
     * When using closed-loop control, the drive motor uses the control output type
     * specified by DriveMotorClosedLoopOutput and any closed-loop
     * SwerveModule#DriveRequestType.
     *
     * \param newDriveMotorGains Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithDriveMotorGains(configs::Slot0Configs newDriveMotorGains)
    {
        this->DriveMotorGains = newDriveMotorGains;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerMotorClosedLoopOutput parameter and returns itself.
     *
     * The closed-loop output type to use for the steer motors.
     *
     * \param newSteerMotorClosedLoopOutput Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithSteerMotorClosedLoopOutput(ClosedLoopOutputType newSteerMotorClosedLoopOutput)
    {
        this->SteerMotorClosedLoopOutput = newSteerMotorClosedLoopOutput;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveMotorClosedLoopOutput parameter and returns itself.
     *
     * The closed-loop output type to use for the drive motors.
     *
     * \param newDriveMotorClosedLoopOutput Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithDriveMotorClosedLoopOutput(ClosedLoopOutputType newDriveMotorClosedLoopOutput)
    {
        this->DriveMotorClosedLoopOutput = newDriveMotorClosedLoopOutput;
        return *this;
    }
    
    /**
     * \brief Modifies the SlipCurrent parameter and returns itself.
     *
     * The maximum amount of stator current the drive motors can apply without
     * slippage.
     *
     * \param newSlipCurrent Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithSlipCurrent(units::current::ampere_t newSlipCurrent)
    {
        this->SlipCurrent = newSlipCurrent;
        return *this;
    }
    
    /**
     * \brief Modifies the SpeedAt12Volts parameter and returns itself.
     *
     * When using open-loop drive control, this specifies the speed at which the
     * robot travels when driven with 12 volts. This is used to approximate the
     * output for a desired velocity. If using closed loop control, this value is
     * ignored.
     *
     * \param newSpeedAt12Volts Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithSpeedAt12Volts(units::velocity::meters_per_second_t newSpeedAt12Volts)
    {
        this->SpeedAt12Volts = newSpeedAt12Volts;
        return *this;
    }
    
    /**
     * \brief Modifies the FeedbackSource parameter and returns itself.
     *
     * Choose how the feedback sensors should be configured.
     * 
     * If the robot does not support Pro, then this should be set to RemoteCANcoder.
     * Otherwise, users have the option to use either FusedCANcoder or SyncCANcoder
     * depending on if there is a risk that the CANcoder can fail in a way to
     * provide "good" data.
     * 
     * If this is set to FusedCANcoder or SyncCANcoder when the steer motor is not
     * Pro-licensed, the device will automatically fall back to RemoteCANcoder and
     * report a UsingProFeatureOnUnlicensedDevice status code.
     *
     * \param newFeedbackSource Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithFeedbackSource(SteerFeedbackType newFeedbackSource)
    {
        this->FeedbackSource = newFeedbackSource;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveMotorInitialConfigs parameter and returns itself.
     *
     * The initial configs used to configure the drive motor of the swerve module.
     * The default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MotorOutputConfigs#NeutralMode (Brake mode, overwritten with
     *   SwerveDrivetrain#ConfigNeutralMode)
     * - configs#MotorOutputConfigs#Inverted
     *   (SwerveModuleConstants#DriveMotorInverted)
     * - configs#Slot0Configs (SwerveModuleConstants#DriveMotorGains)
     * - configs#CurrentLimitsConfigs#StatorCurrentLimit /
     *   configs#TorqueCurrentConfigs#PeakForwardTorqueCurrent /
     *   configs#TorqueCurrentConfigs#PeakReverseTorqueCurrent
     *   (SwerveModuleConstants#SlipCurrent)
     * - configs#CurrentLimitsConfigs#StatorCurrentLimitEnable (Enabled)
     * 
     *
     * \param newDriveMotorInitialConfigs Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithDriveMotorInitialConfigs(configs::TalonFXConfiguration newDriveMotorInitialConfigs)
    {
        this->DriveMotorInitialConfigs = newDriveMotorInitialConfigs;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerMotorInitialConfigs parameter and returns itself.
     *
     * The initial configs used to configure the steer motor of the swerve module.
     * The default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MotorOutputConfigs#NeutralMode (Brake mode)
     * - configs#MotorOutputConfigs#Inverted
     *   (SwerveModuleConstants#SteerMotorInverted)
     * - configs#Slot0Configs (SwerveModuleConstants#SteerMotorGains)
     * - configs#FeedbackConfigs#FeedbackRemoteSensorID
     *   (SwerveModuleConstants#CANcoderId)
     * - configs#FeedbackConfigs#FeedbackSensorSource
     *   (SwerveModuleConstants#FeedbackSource)
     * - configs#FeedbackConfigs#RotorToSensorRatio
     *   (SwerveModuleConstants#SteerMotorGearRatio)
     * - configs#MotionMagicConfigs (Calculated from gear ratios)
     * - configs#ClosedLoopGeneralConfigs#ContinuousWrap (true)
     * 
     *
     * \param newSteerMotorInitialConfigs Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithSteerMotorInitialConfigs(configs::TalonFXConfiguration newSteerMotorInitialConfigs)
    {
        this->SteerMotorInitialConfigs = newSteerMotorInitialConfigs;
        return *this;
    }
    
    /**
     * \brief Modifies the CANcoderInitialConfigs parameter and returns itself.
     *
     * The initial configs used to configure the CANcoder of the swerve module. The
     * default value is the factory-default.
     * 
     * Users may change the initial configuration as they need. Any config that's
     * not referenced in the SwerveModuleConstants class is available to be changed.
     * 
     * The list of configs that will be overwritten is as follows:
     * 
     * - configs#MagnetSensorConfigs#MagnetOffset
     *   (SwerveModuleConstants#CANcoderOffset)
     * 
     *
     * \param newCANcoderInitialConfigs Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithCANcoderInitialConfigs(configs::CANcoderConfiguration newCANcoderInitialConfigs)
    {
        this->CANcoderInitialConfigs = newCANcoderInitialConfigs;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerInertia parameter and returns itself.
     *
     * Simulated azimuthal inertia.
     *
     * \param newSteerInertia Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithSteerInertia(units::moment_of_inertia::kilogram_square_meter_t newSteerInertia)
    {
        this->SteerInertia = newSteerInertia;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveInertia parameter and returns itself.
     *
     * Simulated drive inertia.
     *
     * \param newDriveInertia Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithDriveInertia(units::moment_of_inertia::kilogram_square_meter_t newDriveInertia)
    {
        this->DriveInertia = newDriveInertia;
        return *this;
    }
    
    /**
     * \brief Modifies the SteerFrictionVoltage parameter and returns itself.
     *
     * Simulated steer voltage required to overcome friction.
     *
     * \param newSteerFrictionVoltage Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithSteerFrictionVoltage(units::voltage::volt_t newSteerFrictionVoltage)
    {
        this->SteerFrictionVoltage = newSteerFrictionVoltage;
        return *this;
    }
    
    /**
     * \brief Modifies the DriveFrictionVoltage parameter and returns itself.
     *
     * Simulated drive voltage required to overcome friction.
     *
     * \param newDriveFrictionVoltage Parameter to modify
     * \returns this object
     */
    constexpr SwerveModuleConstantsFactory &WithDriveFrictionVoltage(units::voltage::volt_t newDriveFrictionVoltage)
    {
        this->DriveFrictionVoltage = newDriveFrictionVoltage;
        return *this;
    }
    
    /**
     * \brief Creates the constants for a swerve module with the given properties.
     *
     * \param steerMotorId CAN ID of the steer motor.
     * \param driveMotorId CAN ID of the drive motor.
     * \param cancoderId CAN ID of the CANcoder used for azimuth.
     * \param cancoderOffset Offset of the CANcoder.
     * \param locationX The location of this module's wheels relative to the physical center of the robot in meters along the X axis of the robot.
     * \param locationY The location of this module's wheels relative to the physical center of the robot in meters along the Y axis of the robot.
     * \param driveMotorInverted True if the drive motor is inverted.
     * \param steerMotorInverted True if the steer motor is inverted from the CANcoder.
     * \returns Constants for the swerve module
     */
    constexpr SwerveModuleConstants CreateModuleConstants(
            int steerMotorId,
            int driveMotorId,
            int cancoderId,
            units::angle::turn_t cancoderOffset,
            units::length::meter_t locationX,
            units::length::meter_t locationY,
            bool driveMotorInverted,
            bool steerMotorInverted) const
    {
        return SwerveModuleConstants{}
            .WithSteerMotorId(steerMotorId)
            .WithDriveMotorId(driveMotorId)
            .WithCANcoderId(cancoderId)
            .WithCANcoderOffset(cancoderOffset)
            .WithLocationX(locationX)
            .WithLocationY(locationY)
            .WithDriveMotorInverted(driveMotorInverted)
            .WithSteerMotorInverted(steerMotorInverted)
            .WithDriveMotorGearRatio(DriveMotorGearRatio)
            .WithSteerMotorGearRatio(SteerMotorGearRatio)
            .WithCouplingGearRatio(CouplingGearRatio)
            .WithWheelRadius(WheelRadius)
            .WithSteerMotorGains(SteerMotorGains)
            .WithDriveMotorGains(DriveMotorGains)
            .WithSteerMotorClosedLoopOutput(SteerMotorClosedLoopOutput)
            .WithDriveMotorClosedLoopOutput(DriveMotorClosedLoopOutput)
            .WithSlipCurrent(SlipCurrent)
            .WithSpeedAt12Volts(SpeedAt12Volts)
            .WithFeedbackSource(FeedbackSource)
            .WithDriveMotorInitialConfigs(DriveMotorInitialConfigs)
            .WithSteerMotorInitialConfigs(SteerMotorInitialConfigs)
            .WithCANcoderInitialConfigs(CANcoderInitialConfigs)
            .WithSteerInertia(SteerInertia)
            .WithDriveInertia(DriveInertia)
            .WithSteerFrictionVoltage(SteerFrictionVoltage)
            .WithDriveFrictionVoltage(DriveFrictionVoltage);
    }
};

}
}
}
